@@ -21,6 +21,10 @@
 #include "sscr.h"
 #include "winio.h"
 
+#ifdef MEMWATCH
+  #include "memwatch.h"
+#endif
+
 /* the procedures in this file operate on two data structures: a
  * linked list of aliases and a linked list of "sets". Here a "set" is
  * something that is set by the command "set", i.e., a variable.
@@ -88,7 +92,9 @@ hardwired_t hw[] = {
   { "bandwidthup1", 0, 0, &info.bandwidthupconn, "# upload bandwidth limit per connection (in kB/s)" },
 
   { "announcepongs", 0, 0, 0, "# should receipt of PONG packets be announced?" },
-  { "autopurge", 0, 0, 0, "# automatically delete items from up- and download lists?" },
+  { "autopurge", 0, 0, 0, "# time to delete stopped items from up- and download lists" },
+  { "autopurgeup", 0, 0, 0, "# time to delete stopped items from upload list" },
+  { "autopurgedown", 0, 0, 0, "# time to delete stopped items from download list" },
   { "autoreply", 0, 0, 0, "# a string to send as an automatic reply to incoming user messages" },
   { "connecttimeout", 0, 0, 0, "# timeout after this many seconds while connecting to server" },
   { "cursorfollowsscreen", 0, 0, 0, "# should PgUp and PgDn move the screen, rather than the cursor, on the\n# search result screen?" },
@@ -107,7 +113,9 @@ hardwired_t hw[] = {
   { "noscroll", 0, 0, 0, "# should main screen not scroll to bottom automatically on output?" },
   { "savechannels", 0, 0, 0, "# should your open channels to be saved between sessions?" },
   { "savepass", 0, 0, 0, "# should password be saved in config file?" },
+  { "scrollsize", 0, 0, 0, "# limit on the number of lines in main screen, or 0 for no limit" },
   { "sdefaults", 0, 0, 0, "# default keystrokes for search result screen appearance" },
+  { "sharetypes", 0, 0, 0, "# semicolon-separated list of file extensions to share in addition to mp3 files\n" "# Case insensitive. Use '*' for 'all files'." },
   { "showtoomanyuploads", 0, 0, 0, "# display a message when upload limit is reached?" },
   { "turdsize", 0, 0, 0, "# size of the largest incomplete file which will be auto-removed" }
 };
@@ -198,7 +206,7 @@ int savesets(const char *fn)
       fprintf(f, "%s\n", hw[i].comment);
     v = getrealval(hw[i].nm);
     if (v) {
-      if (hw[i].type==2 && nvar("savepass") != 1) {
+      if (hw[i].type==2 && nvar_default("savepass", 0) == 0) {
 	fprintf(f, "%s=%s\n", hw[i].nm, "?");
       } else {
 	fprintf(f, "%s=%s\n", hw[i].nm, v);
@@ -419,7 +427,13 @@ void delset(const char *nm)
 }
 
 /* lookup set's value as a number, or -1 if not set or a non-number. */
-long nvar(const char *nm)
+long nvar(const char *nm) 
+{
+  return nvar_default(nm, -1);
+}
+
+/* lookup set's value as a number, or default if not set or a non-number. */
+long nvar_default(const char *nm, long deflt)
 {
   sets_t *cur;
   int i;
@@ -427,16 +441,16 @@ long nvar(const char *nm)
   /* check for hardwired sets first */
   for (i=0; i<hwsize; i++) {
     if (!strcasecmp(nm, hw[i].nm) && hw[i].type != 0) {
-      return *hw[i].strp ? strtol(*hw[i].strp, NULL, 10) : -1;
+      return *hw[i].strp ? strtol(*hw[i].strp, NULL, 10) : deflt;
     }
   }
   
   cur = findset(nm);
   if (!cur)
-    return(-1);
+    return(deflt);
   
   if (cur->t)
-    return(-1);
+    return(deflt);
   
   return(cur->intval);
 }
@@ -746,9 +760,14 @@ unsigned char *doalias(alias_t *al, char **tok, int cnt)
       if (atoi(atok[i]+1) > j)
         j = atoi(atok[i]+1);
 
-  if ((cnt-1) < j)
+  if ((cnt-1) < j) {
+    for (i=0; i<acnt; i++) {
+      free(atok[i]);
+    }
+    free(atok);
     return(NULL);
-  
+  }  
+
   for (i=0;atok[i];i++)
   {
     if (*atok[i] == '?' && isdigit(atok[i][1]))
