@@ -31,7 +31,7 @@ typedef struct user_s user_t;
 
 struct hotlist_s 
 {
-  char *nm;
+  char *nm; /* allocated */
   int conn;
   int on;
   struct hotlist_s *next;
@@ -556,6 +556,9 @@ O_NAP_FUNC(dnoprint);
 O_NAP_FUNC(dstop);
 O_NAP_FUNC(dserver);
 O_NAP_FUNC(dgetservers);
+#ifdef MEMWATCH
+O_NAP_FUNC(dmemwatch);
+#endif
 
 char *fixquotes(char *s);
 void upchan(chans_t *);
