@@ -23,6 +23,10 @@
 #include "irc.h"
 #include "alias.h"
 
+#ifdef MEMWATCH
+  #include "memwatch.h"
+#endif
+
 extern info_t info;
 extern chans_t *chanl, *curchan, *recent;
 extern int wmode;
@@ -108,7 +112,7 @@ out_irc_cmd_t outs[] = {
 
 int dccfunc(WINDOW *win, sock_t *m)
 {
-  char *nm = strchr(m->nm, ' ')+1;
+  char *nm = strchr(m->socknm, ' ')+1;
   unsigned char msg[1024];
   int i;
   
@@ -152,12 +156,12 @@ int ldcc(WINDOW *win, sock_t *m)
     return(1);
   }
   
-  msprintf(&nm, "chat %s", strchr(m->nm, ' ')+1);
+  msprintf(&nm, "chat %s", strchr(m->socknm, ' ')+1);
   
   addsock(s, nm, S_R, dccfunc);
   free(nm);
   
-  wp(win, "%s* DCC Connection to %s established%s\n", BRIGHT(CYAN), strchr(m->nm, ' ')+1, WHITE);
+  wp(win, "%s* DCC Connection to %s established%s\n", BRIGHT(CYAN), strchr(m->socknm, ' ')+1, WHITE);
   drw(win);
   
   delsock(m->fd);
@@ -1948,4 +1952,5 @@ O_IRC_FUNC(ichat)
   
   wp(win, "%s* DCC Connection to %s established%s\n", BRIGHT(CYAN), tok[1], WHITE);
   drw(win);
+  free(nm);
 }  
