@@ -3,6 +3,10 @@
  * Please refer to the COPYRIGHT file for more information.
  */
 
+/* the main functions in this file are mp3info and ogginfo. They
+   extract information such as file size, bitrate, sample frequency,
+   and md5 hash from a named file */
+
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -14,6 +18,10 @@
 #include "md5.h"
 #include "mp3s.h"
 
+#ifdef MEMWATCH
+  #include "memwatch.h"
+#endif
+
 /* arrays are indexed using id & bitrate fields from frame header */
 /* 1st row is for MPEG-2 (id=0) Layer III, 2nd for MPEG-1 (id=1) Layer III */
 int bratetab[2][16] = 
@@ -211,7 +219,7 @@ int framesize(hdr_t h)
  *    data. If this false frame header happens to have a bitrate field that
  *    differs from that of the first header, a non-VBR file might be 
  *    incorrectly identified as VBR. As a result, the file will be unneccessarily
- *    fully scanned in gethdr. */
+ *    fully scanned in mp3info. */
 int checkvbr(int f, int startpos, int size, hdr_t firsthdr)
 {
   int stepsize, i, j, isvbr, headerfound, bpf;
@@ -275,20 +283,20 @@ int checkvbr(int f, int startpos, int size, hdr_t firsthdr)
    additional argument md5blocks determines how many 1024-byte blocks
    we hash for the md5 checksum: it is either 292 (=299008 bytes) or
    293 (=300032 bytes).  */
-mhdr_t *gethdr(char *fn, int md5blocks)
+mhdr_t *mp3info(char *fn, int md5blocks)
 {
   hdr_t m;
   mhdr_t *r;
-  int f, i;
+  int f;
   size_t size = 0;
   struct stat st;
   id3v2_t tag;
   char rdt[3];
-  unsigned char md5[16];
   size_t id3size, endtagsize;
   int numframes, bpf, spf;
   int seconds, avgbrate;
   hdr_t h;
+  int res;
 
   f = open(fn, O_RDONLY);
   if (f == -1)
@@ -379,7 +387,7 @@ mhdr_t *gethdr(char *fn, int md5blocks)
       numframes++;
       bpf = framesize(h);
       lseek(f, bpf-4, SEEK_CUR); /* seek to next frame header */
-    } /* end while gethdr */
+    } /* end while getrhdr */
     /* at this point we have reached EOF or an invalid header */
     if (numframes == 0) { /* no valid headers found */
       close(f);
@@ -403,31 +411,147 @@ mhdr_t *gethdr(char *fn, int md5blocks)
     r->bitrate = bratetab[m.id][m.brate];
     r->len = (r->sz1/r->bitrate)/125;
   }
-    
 
-  /* finally, calculate md5 hash. According to drscholl's napster
-  spec, the official client hashes the first 299008 bytes, or 300032
-  bytes, *not* 300000 bytes as nap used to do. I did three random spot
-  checks, and in all three cases, the hash matched at 300032. Thus, we
-  now use 300032 bytes, hoping that this will be more conforming to
-  general usage. By the way, 299008 = 1024 * 292, and 300032 = 1024 *
-  293. */
-
-  /* rewind to beginning of mp3 header (frame synch) */
+  /* finally, seek to end of ID3 header and calculate md5 hash */
   lseek(f, id3size, SEEK_SET);
-  if (md5_stream_blocks(f, md5blocks, md5)) {
+  res = md5hash(r->check, f, md5blocks);
+  if (res) {
     free(r);
     close(f);
+    return(NULL);
+  }
+
+  close(f);
+  return(r);
+}
+
+/* support for ogg vorbis files: get standard information (bitrate,
+   frequency, length in seconds, size in bytes, and checksum) from the
+   names ogg vorbis file. Return NULL if not recognized as a valid ogg
+   vorbis file. */
+mhdr_t *ogginfo(char *fn, int md5blocks)
+{
+  /* note: since ogg uses variable bitrate, to keep things simple, we
+     use the "nominal", rather than the "actual" bitrate of the ogg
+     stream. This information can be read from the first header
+     packet.  Otherwise we'd have to scan the entire file. We also
+     only give a rough estimate of the song length, based on the file
+     size. Sorry. Scanning an ogg vorbis file for that information
+     would require ogg and vorbis parsers, which is too much overhead
+     for us here. */
+
+  /* we make some simple assumptions: the file contains an ogg stream
+     revision 0, whose first page contains precisely one package,
+     which is a vorbis "1" header. All other kinds of ogg streams will
+     fail here. Information on ogg streams comes from
+     http://www.xiph.org/ogg/vorbis/doc/framing.html, information on
+     vorbis "1" header comes from vorbis/lib/info.c in the vorbis
+     distribution. */
+
+  unsigned char buf[0x58];
+  int n;
+  int f;
+  mhdr_t *r;
+  struct stat st;
+
+  /* get file stats and read first 58 bytes */
+  f = open(fn, O_RDONLY);
+  if (f == -1) {
+    return NULL;
+  }
+  fstat(f, &st);
+  n = read(f, buf, 58);
+  close(f);
+
+  if (n<58) {
+    return NULL;
+  }
+  
+  /* check that this is a first page of a logical ogg stream. We don't
+     bother checking the CRC checksum */
+  if (strncmp(&buf[0], "OggS", 4) != 0) {        /* capture pattern */
+    return NULL;
+  }
+  if (buf[4] != 0) {                             /* stream structure version */
+    return NULL;
+  }
+  if ((buf[5] & 3) != 2) {                       /* ogg header type flag */
+    return NULL;
+  }
+  if (buf[18] || buf[19] || buf[20] || buf[21]) {/* page sequence no. */
+    return NULL;                                 
+  }
+  if (buf[26] != 1 || buf[27] != 30) {           /* segment table */
     return NULL;
   }
 
+  /* check that this is a vorbis "1" header */
+  if (buf[28] != 1) {                            /* vorbis packet type */
+    return NULL;
+  }
+  if (strncmp(&buf[29], "vorbis", 6) != 0) {     /* vorbis packet identifier */
+    return NULL;
+  }
+  if (buf[35] || buf[36] || buf[37] || buf[38]) {/* vorbis version */
+    return NULL;                                 
+  }
+  if ((buf[57] & 1) != 1) {                        /* EOP check */
+    return NULL;
+  }
+  
+  /* seems we found an okay vorbis packet, now read the data. We
+     ignore the "upper" and "lower" bitrates and go straight for the
+     nominal bitrate. */
+
+  r = (mhdr_t *)malloc(sizeof(mhdr_t));
+  if (!r) {   
+    return NULL;   /* out of memory; fail silently */
+  }
+
+  /* channels = buf[39]; */
+  r->freq    = buf[40] | (buf[41]<<8) | (buf[42]<<16) | (buf[43]<<24);
+  r->bitrate = buf[48] | (buf[49]<<8) | (buf[50]<<16) | (buf[51]<<24);
+  r->sz      = st.st_size;
+  r->sz1     = st.st_size;
+
+  /* we have to estimate the length, based on bitrate and file size */
+  r->len     = r->sz * 8 / r->bitrate;
+
+  /* convert from bps to kbps */
+  r->bitrate /= 1000;  
+
+  /* calculate checksum */
+  lseek(f, 0, SEEK_SET);
+  md5hash(r->check, f, md5blocks);
+
+  return r;
+}
+
+/* calculate md5 hash of the given stream, starting from the current
+   file position. md5bocks is the number of blocks to hash. Write to
+   check. Return -1 on error, else 0. */
+
+int md5hash(char check[33], int fd, int md5blocks) {
+  unsigned char md5[16];
+  int i;
+
+  /* According to drscholl's napster spec, the official client hashes
+     the first 299008 bytes, or 300032 bytes, *not* 300000 bytes as
+     nap used to do. I did three random spot checks, and in all three
+     cases, the hash matched at 300032. Thus, we now use 300032 bytes,
+     hoping that this will be more conforming to general usage. By the
+     way, 299008 = 1024 * 292, and 300032 = 1024 * 293. -PS */
+
+  if (md5_stream_blocks(fd, md5blocks, md5)) {
+    return(-1);
+  }
+
   for (i=0;i<16;i++)
-    sprintf(r->check, "%s%02x", r->check, md5[i]);
+    sprintf(check+2*i, "%02x", md5[i]);
   
-  r->check[32] = '\0';
+  /*  check[32] = '\0'; */
   
-  close(f);
-  return(r);
+  return(0);
 }
 
 /* the next function really belongs in md5.c, but since that file is
