@@ -44,7 +44,11 @@
 #include "winio.h"
 
 #ifdef MCURSES
-#include "wind.h"
+  #include "wind.h"
+#endif
+
+#ifdef MEMWATCH
+  #include "memwatch.h"
 #endif
 
 /* standard napster servers */
@@ -52,8 +56,6 @@
 extern int ircin;
 extern int ircsock;
 extern int ups[], speeds[];
-extern scrls_t *mscroll, *scur;
-extern chans_t *recent;
 extern WINDOW *wchan, *winput;
 extern char cbuf[];
 extern upload_t *up;
@@ -70,7 +72,7 @@ int dresize = 0;
 
 void doquit()
 {
-  scur = NULL;
+  scrollbottom();
   if (wchan)
   {
     dscr(wchan);
@@ -82,7 +84,7 @@ void doquit()
     close(ircsock);
     ircsock = 0;
   }
-  if (nvar("savechannels") == 1) {
+  if (nvar_default("savechannels", 0)) {
     char *tmp = NULL, *fn;
     msprintf(&tmp, CHANNELFILE, info.user);
     fn = home_file(tmp);
@@ -514,6 +516,9 @@ int set_defaults_interactive(char *fn) {
     }
   }
 
+  free(ans);
+  ans = NULL;
+
   if (!getval("connection")) {
     if (!info.daemon) {
       wp(NULL, "
@@ -671,7 +676,7 @@ static struct option longopts[] = {
   {"incomplete",  1, 0, 'I'},
   {"dataport",    1, 0, 'P'},
   {"connection",  1, 0, 'C'},
-  {"maxuploads",    1, 0, 'M'},
+  {"maxuploads",  1, 0, 'M'},
   {"option",      1, 0, 'o'},
   {0, 0, 0, 0}
 };
@@ -680,34 +685,34 @@ void phelp(char *nm)
 {
   printf("Usage: %s [options]\n", nm);
   printf("Options:\n");
-  printf("-h, --help         - print this help message\n");
-  printf("-v, --version      - print version info and exit\n");
-  printf("-b, --build        - build library of your current mp3s to send to server\n");
-  printf("-B, --build-only   - build library and exit\n");
-  printf("-N, --nobuild      - do not build library, even if it is out of date\n");
-  printf("-m, --create       - create a new account with the napster server (obsolete)\n");
-  printf("-r, --reconnect    - keep reconnecting until server connection established\n");
-  printf("-a, --autorestart  - automatically reconnect when connection to server lost\n");
-  printf("-q, --daemon       - run without user interface; file sharing only\n");
-  printf("-t, --notitle      - do not display the title bar (fixes messed-up displays)\n");
-  printf("-l, --nxterm       - try using a terminal which is compatible with most\n"
-	 "                     systems (fixes some messed-up displays)\n");
-  printf("-T, --transparent  - use the terminal's default background instead of black\n");
-  printf("-f fn, --config fn - specifies the config file to use (default $HOME/"CONFIGFILE")\n");
-  printf("-x fn, --log fn    - log all transfers to a specific filename\n");
-  printf("-g fn, --logall fn - log everything to a specific filename\n");
-  printf("-s sv, --server sv - select a specific server (multiple -s opts possible)\n");
-  printf("-d n, --debug n    - set debug level\n");
-  printf("-u str, --user str     - specify napster username\n");
-  printf("-p str, --pass str     - specify user's password\n");
-  printf("-e str, --email str    - specify user's email address\n");
-  printf("-U dir, --upload dir   - specify upload directory (multiple -U opts possible)\n");
-  printf("-D dir, --download dir - specify download directory\n");
-  printf("-I dir, --incomplete dir - specify directory for incomplete files\n");
-  printf("-P n-m, --dataport n-m - specify port(s) to use for incoming upload requests\n");
-  printf("-C n, --connection n   - specify connection speed number (see README)\n");
-  printf("-M n, --maxuploads n   - specify maximum number of simultaneous uploads\n");
-  printf("-o var=value, --option var=value  - set user variable\n");
+  printf("-h, --help             - print this help message\n");
+  printf("-v, --version          - print version info and exit\n");
+  printf("-b, --build            - build library of shared files to send to server\n");
+  printf("-B, --build-only       - build library and exit\n");
+  printf("-N, --nobuild          - do not build library, even if it is out of date\n");
+  printf("-m, --create           - create new account on server (obsolete)\n");
+  printf("-r, --reconnect        - keep reconnecting until server connection established\n");
+  printf("-a, --autorestart      - automatically reconnect when connection to server lost\n");
+  printf("-q, --daemon           - run without user interface; file sharing only\n");
+  printf("-t, --notitle          - do not display title bar (fixes messed-up displays)\n");
+  printf("-l, --nxterm           - try using a terminal which is compatible with most\n"
+	 "                         systems (fixes some messed-up displays)\n");
+  printf("-T, --transparent      - use terminal's default background instead of black\n");
+  printf("-f, --config fn        - config file to use (default $HOME/"CONFIGFILE")\n");
+  printf("-x, --log fn           - log all transfers to a specific filename\n");
+  printf("-g, --logall fn        - log everything to a specific filename\n");
+  printf("-s, --server sv        - select a specific server (multiple -s opts possible)\n");
+  printf("-d, --debug n          - set debug level\n");
+  printf("-u, --user str         - specify napster username\n");
+  printf("-p, --pass str         - specify user's password\n");
+  printf("-e, --email str        - specify user's email address\n");
+  printf("-U, --upload dir       - specify upload directory (multiple -U opts possible)\n");
+  printf("-D, --download dir     - specify download directory\n");
+  printf("-I, --incomplete dir   - specify directory for incomplete files\n");
+  printf("-P, --dataport n-m     - specify port(s) to use for incoming upload requests\n");
+  printf("-C, --connection n     - specify connection speed number (see README)\n");
+  printf("-M, --maxuploads n     - specify maximum number of simultaneous uploads\n");
+  printf("-o, --option var=value - set user variable\n");
 }
 
 void pversion() {
@@ -744,6 +749,7 @@ void dopts(int argc, char **argv)
   info.bandwidthup = -1;
   info.bandwidthdownconn = -1;
   info.bandwidthupconn = -1;
+  info.sflag = 0;
 
   while ((c = getopt_long(argc, argv, "h?vbBNmrqatlTf:x:g:s:d:u:p:e:U:D:P:C:M:o:", 
 			  longopts, NULL)) != -1)
@@ -788,7 +794,7 @@ void dopts(int argc, char **argv)
 	strcat(info.serverlist, ";");
 	strcat(info.serverlist, optarg);
       }
-      chset("nometa", "1");
+      info.sflag = 1;
       break;
     case 'o':
       p = strchr(optarg, '=');
@@ -956,7 +962,7 @@ int setup_configdir(void) {
     move_file(tmp1, tmp2);
     free(tmp1);
     free(tmp2);
-
+    free(configdir);
     return 0;
   }
 
@@ -1083,12 +1089,17 @@ int main(int argc, char *argv[])
   FILE *fl;
   char *fn;
   const char *errmsg;
-  char *libraryfile;
+  char *libraryfile, *aliasfile, *handlerfile;
+  char *ans, *serial;
+  FILE *nfile;
+  char *nfilename;
+
+  /* set up some signal handlers (more below) */
+  signal(SIGCHLD, SIG_IGN);
+  signal(SIGPIPE, SIG_IGN);
 
   /* initialize main screen's internal state */
   cbuf[0] = 0;
-  mscroll = scur = NULL;
-  recent = NULL;
 
   screen = MAIN_SCREEN;
   
@@ -1116,14 +1127,16 @@ int main(int argc, char *argv[])
   
   if (info.build != 2) {
     r = set_defaults_interactive(fn);
-    if (r==-1)
+    if (r==-1) {
+      free(fn);
       return(1);
+    }
   }
 
   free(fn);
 
   if (info.shared_filename) {
-    libraryfile = info.shared_filename;
+    libraryfile = strdup(info.shared_filename);
   } else {
     libraryfile = home_file(LIBRARYFILE);
   }
@@ -1134,6 +1147,7 @@ int main(int argc, char *argv[])
 	wp(NULL, "Building library...\n");
 	if (buildflist(libraryfile, info.up) == -1) {
 	  wp(NULL, "There was an error building your library.\n");
+	  free(libraryfile);
 	  return(1);
 	}
       } else if (!up_to_date(libraryfile, info.up))
@@ -1141,6 +1155,7 @@ int main(int argc, char *argv[])
 	  wp(NULL, "Your library is not up-to-date. Rebuilding...\n");
 	  if (buildflist(libraryfile, info.up) == -1) {
 	    wp(NULL, "There was an error building your library.\n");
+	    free(libraryfile);
 	    return(1);
 	  }
 	}
@@ -1154,32 +1169,50 @@ int main(int argc, char *argv[])
   {
     wp(NULL, "Error reading %s: %s\nTry running \"nap -b\".\n", \
        libraryfile, strerror(errno));
+    free(libraryfile);
     return(1);
   }
+  free(libraryfile);
   fclose(fl);
 
   /* check for news on the client */
-  if (nvar("nonews")!=1 && !info.daemon) {
+  if (nvar_default("nonews", 0)==0 && !info.daemon) {
     int timeout;
 
-    if (getval("newstimeout"))
-      timeout = nvar("newstimeout");
-    else
-      timeout = NEWSTIMEOUT;
+    timeout = nvar_default("newstimeout", NEWSTIMEOUT);
+
     wp(NULL, "Checking for new releases of nap... ");
-    r = checknv(1, timeout, &errmsg);
-    if (r==1) {
-      wp(NULL, "Press return to continue.");
-      getline(stdin);
+    r = checknv(1, timeout, &errmsg, &serial);
+    if (r==1 && serial) {
+      wp(NULL, "(o)kay, (q)uit, (d)on't prompt again: ");
+      ans = getline(stdin);
+      if (ans[0]=='q') {
+	exit(0);
+      } else if (ans[0]=='d') {
+	nfilename = home_file(OLDNEWSFILE);
+	nfile = fopen(nfilename, "w");
+	free(nfilename);
+	if (nfile) {
+	  fprintf(nfile, "%s\n", serial);
+	  fclose(nfile);
+	}
+      }
+      free(ans);
+    } else if (r==1) {
+      wp(NULL, "Press return to continue.\n");
+      ans = getline(stdin);
+      free(ans);
     } else if (r==0) {
       wp(NULL, "none found.\n");
+    } else if (r==2) {  /* old news, don't prompt */
+      /* do nothing */
     } else {
       wp(NULL, "failed.\n");
     }
   }
   
   /* get server list from metaserver */
-  if (nvar("nometa")!=1) {
+  if (!info.sflag && nvar_default("nometa", 0)==0) {
     char *url;
     int timeout;
     const char *errmsg;
@@ -1188,7 +1221,7 @@ int main(int argc, char *argv[])
     if (!url) 
       url = METASERVER;
     
-    timeout = nvar("metatimeout");
+    timeout = nvar_default("metatimeout", METATIMEOUT);
     if (timeout<0)
       timeout = METATIMEOUT;
     
@@ -1300,15 +1333,6 @@ int main(int argc, char *argv[])
       wp(NULL, "Connected to %s:%i\n", inet_ntoa(dst.sin_addr), ntohs(dst.sin_port));
   }
 
-  signal(SIGCHLD, SIG_IGN);
-  signal(SIGPIPE, SIG_IGN);
-  signal(SIGWINCH, tresize);
-  if (!info.daemon)
-    signal(SIGINT, sigint);
-#ifdef MCURSES
-  signal(SIGCONT, tresize);
-#endif
-
   pipe(ipcs);
   
   addsock(ipcs[0], "ipc", S_R, inipc);
@@ -1316,12 +1340,28 @@ int main(int argc, char *argv[])
   if (!info.daemon)
     addsock(0, "input", S_R, input);
   initwin(info.nxterm);
+
+  signal(SIGWINCH, tresize);
+#ifdef MCURSES
+  signal(SIGCONT, tresize);
+#endif
+
+  if (!info.daemon) {
+    signal(SIGINT, sigint);
+  }
+
   t = glistn(info.user);
   checkhotlist(s, t);
   free(t);
-  loadaliases(home_file(ALIASFILE));
-  loadhandlers(home_file(HANDLERFILE));
-  if (nvar("savechannels") == 1) {
+
+  aliasfile = home_file(ALIASFILE);
+  handlerfile = home_file(HANDLERFILE);
+  loadaliases(aliasfile);
+  loadhandlers(handlerfile);
+  free(aliasfile);
+  free(handlerfile);
+
+  if (nvar_default("savechannels", 0)) {
     char *tmp = NULL, *fn;
     msprintf(&tmp, CHANNELFILE, info.user);
     fn = home_file(tmp);
@@ -1425,6 +1465,7 @@ int buildflist(char *sd, char *path)
   FILE *f;
   char *t, *n;
   dev_inode_t *di_list=NULL, *e;
+  char *sharetypes = getval("sharetypes");
 
   f = fopen(sd, "w");
   if (f == NULL)
@@ -1435,6 +1476,7 @@ int buildflist(char *sd, char *path)
   
   fprintf(f, NAP_LIBRARY_HEADER "\n");
   fprintf(f, "PATH=%s\n", path ? path : "");
+  fprintf(f, "SHARETYPES=%s\n", sharetypes ? sharetypes : "");
 
   /* note: if path==NULL, we just generate an empty file */
 
@@ -1520,7 +1562,7 @@ void addfile(char *fn, FILE *f, dev_inode_t **di_list_p)
 
   } else if (S_ISREG(st.st_mode)) { /* is it a regular file? */
 
-    m = gethdr(fn, 293);
+    m = filestats(fn, 293);
     if (m) {
       fprintf(f, "\"%s\" %s-%i %i %i %i %i\n", fn, m->check, m->sz1, m->sz, m->bitrate, m->freq, m->len);
       free(m);
@@ -1530,6 +1572,79 @@ void addfile(char *fn, FILE *f, dev_inode_t **di_list_p)
   }
 }
 
+/* decide whether a given file is shared, and if yes, get the file
+   stats (checksum, size, bitrate etc). A file is shared if: either it
+   is a valid mp3 file, or its file extension is in the sharetypes
+   list. Thus, the sharetypes variable only needs to list file
+   extensions other than mp3 - valid mp3 files are shared irrespective
+   of their extension. */
+
+mhdr_t *filestats(char *fn, int md5blocks) {
+  mhdr_t *m;
+  char *p, *q;
+  int plen, fnlen;
+  int r;
+  int f;
+  struct stat st;
+  char *sharetypes;
+
+  /* first, check if this is a valid mp3 file (irrespective of extension) */
+  m = mp3info(fn, md5blocks);
+  if (m)
+    return m;
+
+  /* next, check if it's a valid ogg vorbis file (irrespective of extension) */
+  m = ogginfo(fn, md5blocks);
+  if (m)
+    return m;
+
+  /* not a valid music file - check to see if it is shared by its file
+     extension */
+  sharetypes = getval("sharetypes");
+  if (!sharetypes) {
+    return NULL;
+  }
+
+  fnlen = strlen(fn);
+
+  p = sharetypes;
+  for (p=sharetypes; p != NULL+1; p = strchr(p, ';')+1) {
+    q = strchr(p, ';');
+    plen = (q) ? q-p : strlen(p);
+    if (plen == 1 && *p=='*')
+      goto matches;
+    if (fnlen >= plen+1 && fn[fnlen-plen-1]=='.' && strncasecmp(fn+fnlen-plen, p, plen)==0)
+      goto matches;
+  }
+  return NULL;
+
+ matches:
+  /* try to open file */
+  f = open(fn, O_RDONLY);
+  if (f == -1)
+    return(NULL);
+  
+  /* get file attributes, e.g. file size */
+  fstat(f, &st);
+
+  m = (mhdr_t *)malloc(sizeof(mhdr_t));
+  /* not a music file: set freq, len, and bitrate to 0 */
+  m->freq = 0;
+  m->len = 0;
+  m->bitrate = 0;
+  m->sz = st.st_size;
+  m->sz1 = st.st_size;
+  r = md5hash(m->check, f, md5blocks);
+  if (r) {
+    free(m);
+    close(f);
+    return(NULL);
+  }
+
+  close(f);
+  return(m);
+}
+
 /* send to the server the list of files that we are sharing. SD is the
    library filename.  */
 int lfiles(int s, char *sd)
@@ -1541,8 +1656,9 @@ int lfiles(int s, char *sd)
   f = fopen(sd, "r");
   if (f == NULL)
   {
-    wp(NULL, "Error reading %s: %s\nTry running \"nap -b\".\n", sd,
+    wp(wchan, "Error reading %s: %s\nTry running \"nap -b\".\n", sd,
        strerror(errno));
+    drw(wchan);
     return(-1);
   }
   
@@ -1558,9 +1674,12 @@ int lfiles(int s, char *sd)
     for (i=0; r[i]; i++)
       if (r[i] == '/')
         r[i] = '\\';
-    if (sendpack(s, F_SFILE, "%s", r) == -1)  
+    if (sendpack(s, F_SFILE, "%s", r) == -1 || lpbrk)  
     {
-      wp(NULL, "Connection reset by peer\n");
+      if (!lpbrk) {
+	wp(wchan, "Error while sharing files: %s\n", strerror(errno));
+	drw(wchan);
+      }
       free(r);
       return(-1);
     }
@@ -1583,6 +1702,7 @@ int up_to_date(char *sd, char *path) {
   time_t mtime;
   dev_inode_t *di_list, *e;
   int ret, res;
+  char *sharetypes = getval("sharetypes");
 
   if (!sd) {
     return 0;
@@ -1613,7 +1733,21 @@ int up_to_date(char *sd, char *path) {
     return 0;
   }
 
-  if (strncmp(r, "PATH=", 5) || strcmp(r+5, path ? path : "")) { 
+  if (strncmp(r, "PATH=", 5) || strcasecmp(r+5, path ? path : "")) { 
+    fclose(f);
+    free(r);
+    return 0;
+  }
+  
+  free(r);
+
+  r = getline(f);
+  if (!r) {
+    fclose(f);
+    return 0;
+  }
+
+  if (strncmp(r, "SHARETYPES=", 11) || strcmp(r+11, sharetypes ? sharetypes : "")) { 
     fclose(f);
     free(r);
     return 0;
@@ -1772,62 +1906,52 @@ unsigned long int swapl(unsigned long int x) {
 }
 
 /* just like connect(2), except that we time out after t seconds. No
-   timeout if t=0. If a timeout occurred, we set errno to ETIMEDOUT.
-   If we return -1, then errno is either set to a valid error number,
-   or to 255 if there was an unknown error. */
-int connect_t(int fd, struct sockaddr *serv_addr, int addlen, int t) 
+   timeout if t=0. Return value is -1 on error, else 0. If a timeout
+   occurred, we set errno to ETIMEDOUT. */
+
+int connect_t(int fd, struct sockaddr *serv_addr, int addlen, int t)
 {
-  int status;
+  long flags;
   int r;
-  pid_t pid;
-  
+  fd_set fds;
+  struct timeval sec;
+
   if (t==0) {
     return connect(fd, serv_addr, addlen);
   }
 
-  /* we make the connection in a child process, and kill it if it
-     doesn't finish in time. The only feedback we get from the child
-     is 8 bits of its exit status; we use this to pass errno back to
-     the parent process. In case errno>255 (can this ever happen?) the
-     child returns 255.  */
+  /* set to non-blocking */
+  flags = fcntl(fd, F_GETFL);
 
-  pid = fork();
-  if (!pid) { /* child process */
-    r = connect(fd, serv_addr, addlen);
-    if (r == -1) {
-      exit(errno>255 ? 255 : errno);
-    } else {
-      exit(0);
-    }
+  fcntl(fd, F_SETFL, flags | O_NONBLOCK);
+
+  r = connect(fd, serv_addr, addlen);
+
+  /* restore original flags */
+  fcntl(fd, F_SETFL, flags);
+
+  if (r == -1 && errno != EINPROGRESS) {
+    return -1;
   }
+
+  /* wait up to timeout seconds until this fd becomes writeable */
+
+  FD_ZERO(&fds);
+  FD_SET(fd, &fds);
+  sec.tv_sec = t;
+  sec.tv_usec = 0;
+  r = select(fd+1, NULL, &fds, NULL, &sec);
   
-  /* parent process: sleep for t seconds, but make sure we are woken
-     up if the child terminates */
-  signal(SIGCHLD, noop);
-  if (t>0)
-    t = sleep(t);
-  signal(SIGCHLD, SIG_IGN);
-
-  r = waitpid(pid, &status, WNOHANG);
-
-  if (r==-1) {
-    kill(pid, SIGTERM);
-    return(-1);
-  } else if (r==0) {
-    kill(pid, SIGTERM);
-    waitpid(pid, &status, 0);      /* wait for the child to exit */
-    if (t>0) {
-      errno = EINTR;
-    } else {
-      errno = ETIMEDOUT;
-    }
-    return(-1);
-  } else if (WEXITSTATUS(status)) {
-    errno = WEXITSTATUS(status);
-    return(-1);
-  } else {
-    return(0);
+  if (r == -1) {  /* error */
+    return -1;
   }
+
+  if (r == 0) {   /* timeout */
+    errno = ETIMEDOUT;
+    return -1;
+  }
+
+  return 0;
 }
 
 /* make connection to a napster server or meta server. If "host"
@@ -1843,10 +1967,7 @@ int conn(char *host, unsigned short port)
   int i, r;
   int timeout;
 
-  if (getval("connecttimeout"))
-    timeout = nvar("connecttimeout");
-  else
-    timeout = CONNECTTIMEOUT;
+  timeout = nvar_default("connecttimeout", CONNECTTIMEOUT);
 
   if (!strchr(host, ':'))
   {
@@ -1870,8 +1991,12 @@ int conn(char *host, unsigned short port)
     }
     
     if (connect_t(s, (struct sockaddr *)&dst, sizeof(dst), timeout) == -1) {
-      wp(wchan, "Error connecting to %s: %s\n", host, errno!=255 ? strerror(errno) : "unknown error");
-      drw(wchan);
+      /* print error message only if not caused by a user interrupt */
+
+      if (!lpbrk) {
+	wp(wchan, "Error connecting to %s: %s\n", host, strerror(errno));
+	drw(wchan);
+      }
       close(s);
       return(-1);
     }
@@ -1941,8 +2066,10 @@ int conn(char *host, unsigned short port)
 
   if (connect_t(s, (struct sockaddr *)&dst, sizeof(dst), timeout) == -1)
   {
-    wp(wchan, "Error connecting to %s:%d: %s\n", buf, port, errno!=255 ? strerror(errno) : "unknown error");
-    drw(wchan);
+    if (!lpbrk) {
+      wp(wchan, "Error connecting to %s:%d: %s\n", buf, port, strerror(errno));
+      drw(wchan);
+    }
     close(s);
     free(buf);
     return(-1);
@@ -1958,20 +2085,29 @@ int login(int s, char *user, char *pass, int data, char *client, int conn, char
   phead_t *rcv;
   int r;
   char *libraryfile;
+  int timeout;
+
+  timeout = nvar_default("connecttimeout", CONNECTTIMEOUT);
+
+  if (timeout==0) {
+    timeout=-1;
+  }
 
   msprintf(&t1, "nap v%s", client);
   r = sendpack(s, F_LOGIN, "%s %s %i \"%s\" %i", user, pass, data, t1, conn);
   if (r<0) {
-    wp(wchan, "Error while sending data: %s\n", strerror(errno));
+    wp(wchan, "Error: %s\n", strerror(errno));
     drw(wchan);
+    free(t1);
     return(-1);
   }
 
   free(t1);
   t1 = NULL;
-  r = recvpack(s, &t1, &rcv);
-  while (r == -2)
-    r = recvpack(s, &t1, &rcv);
+  r = recvpack_t(s, &t1, &rcv, timeout);
+  while (r == -2) {
+    r = recvpack_t(s, &t1, &rcv, timeout);
+  }
 
   if (r == -1) {
     wp(wchan, "Error while receiving data: %s\n", strerror(errno));
@@ -1981,6 +2117,10 @@ int login(int s, char *user, char *pass, int data, char *client, int conn, char
     wp(wchan, "Error: unexpected end of file from server\n");
     drw(wchan);
     return(-1);
+  } else if (r == -4) {
+    wp(wchan, "Error: Connection timed out\n");
+    drw(wchan);
+    return(-1);
   }
 
   if (rcv->op == F_LOGERROR)
@@ -2000,12 +2140,15 @@ int login(int s, char *user, char *pass, int data, char *client, int conn, char
   free(rcv);
   
   if (info.shared_filename) {
-    libraryfile = info.shared_filename;
+    libraryfile = strdup(info.shared_filename);
   } else {
     libraryfile = home_file(LIBRARYFILE);
   }
 
-  return(lfiles(s, libraryfile));
+  r = lfiles(s, libraryfile);
+
+  free(libraryfile);
+  return r;
 }
 
 int makeact(int s, char *user, char *pass, int data, char *client, int conn, char *email)
@@ -2094,12 +2237,15 @@ int makeact(int s, char *user, char *pass, int data, char *client, int conn, cha
   free(rcv);
   
   if (info.shared_filename) {
-    libraryfile = info.shared_filename;
+    libraryfile = strdup(info.shared_filename);
   } else {
     libraryfile = home_file(LIBRARYFILE);
   }
 
-  return(lfiles(s, libraryfile));
+  r = lfiles(s, libraryfile);
+
+  free(libraryfile);
+  return r;
 }
 
 /* calculate the name of the hotlist file */
@@ -2160,22 +2306,32 @@ void checkhotlist(int s, char *fn)
    and we ignore lines whose keyword we don't know. Note: this
    function may be executed in a child process, since it does not
    update any state. Return 1 if there was news, 0 if there was none,
-   and -1 on error. In case of -1, the variable *errmsg is set to a
-   (static) error message. If fd is 1, do some special formatting,
-   write in black and white, and send the output through wp() for the
-   log file's benefit. */
-int checknv(int fd, int timeout, const char **errmsg)
+   and -1 on error. Return 2 if there was news, but serial!=NULL and
+   we have seen these very news before. In case of -1, the variable
+   *errmsg is set to a (static) error message. If fd is 1, do some
+   special formatting, write in black and white, and send the output
+   through wp() for the log file's benefit. If we return 1 (there is
+   new news) and serial!=NULL and the news had a serial identifier,
+   that identifier is returned in *serial. */
+int checknv(int fd, int timeout, const char **errmsg, char **serial)
 {
-  FILE *f;
-  int relevant, news;
-  char *b;
+  FILE *f, *nfile;
+  int relevant, news, oldnews;
+  char *b, *c;
+  char *nfilename;
+
+  if (serial) {
+    *serial = NULL;
+  }
 
   f = open_url(NEWSURL, timeout, errmsg);
-  if (!f)
+  if (!f) {
     return -1;
+  }
 
   relevant = 0;
   news = 0;
+  oldnews = 0;
   
   /* parse each line of the file */
   while ((b = getline(f)) != NULL) {
@@ -2183,6 +2339,28 @@ int checknv(int fd, int timeout, const char **errmsg)
       free(b);
       continue;
     }
+    if (serial && !strncmp(b, "SERIAL ", 7)) {
+      /* check to see if this version of the news is old */
+      nfilename = home_file(OLDNEWSFILE);
+      nfile = fopen(nfilename, "r");
+      free(nfilename);
+      if (nfile) {
+	c = getline(nfile);
+	fclose(nfile);
+	if (c && !strcmp(b+7, c)) {
+	  oldnews = 1;
+	}
+	if (c) {
+	  free(c);
+	}
+      }
+      if (*serial) {
+	free(*serial);
+      }
+      *serial = strdup(b+7);
+      free(b);
+      continue;
+    }
     if (!strncmp(b, "VERSION ", 8) && !strcmp(b+8, VERSION)) {
       relevant = 1;      /* news for this client */
       free(b);
@@ -2211,12 +2389,26 @@ int checknv(int fd, int timeout, const char **errmsg)
       free(b);
       continue;
     }
+    free(b);
   }
   if (news && fd==1) {
     wp(NULL, "**********************************************************************\n\n");
   }
   fclose(f);
-  return news;
+
+  if (news && !oldnews) {
+    return 1;
+  }
+
+  if (serial && *serial) {
+    free(*serial);
+  }
+
+  if (news) {
+    return 2;
+  }
+
+  return 0;
 } 
 
 /* save channels from channel list CL to file FN. Return -1 on i/o
@@ -2270,13 +2462,14 @@ chans_t *findchan(chans_t *h, char *chan)
 {
   chans_t *cur;
 
-  for (cur=h;;cur=cur->next)
-  {
-    if (!cur)
-      return(NULL);
-    if (!strcasecmp(cur->nm, chan))
-      return(cur);
+  list_find(cur, h, !strcasecmp(cur->nm, chan));
+  if (cur)
+    return cur;
+
+  if (chan[0]!='#') {
+    list_find(cur, h, cur->nm[0]=='#' && !strcasecmp(cur->nm+1, chan));
   }
+  return(cur);
 }
 
 chans_t *findquery(chans_t *h, char *chan)
@@ -2293,8 +2486,9 @@ chans_t *findquery(chans_t *h, char *chan)
 }
 
 /* for sending a packet to the napster server, whose file descriptor
-   is s. Return -1 on error (with errno not necessarily set), or else
-   the number of bytes written. */
+   is s. Return -1 on error (with errno set), or else the number of
+   bytes written. If s==-1, we print an error message and set errno to
+   EINVAL. */
 int sendpack(int s, int op, const char *fmt, ...)
 {
   char *data;
@@ -2307,6 +2501,7 @@ int sendpack(int s, int op, const char *fmt, ...)
   {
     wp(wchan, "%s* Not connected to the server%s\n", RED, WHITE);
     drw(wchan);
+    errno = EINVAL;
     return(-1);
   }
   
@@ -2358,7 +2553,7 @@ int ssock(int s, const char *fmt, ...)
   if (nvar("debug") == 2)
   {
     sock_t *sk = findsockfd(s);
-    wp(wchan, ""DARK GREEN"--> [to %d=%s] <%s>"WHITE"\n", s, sk ? sk->nm : "?", quote(data));
+    wp(wchan, ""DARK GREEN"--> [to %d=%s] <%s>"WHITE"\n", s, sk ? sk->socknm : "?", quote(data));
     drw(wchan);
   }
 
@@ -2374,7 +2569,14 @@ int ssock(int s, const char *fmt, ...)
    unexpected end-of-file. Return -2 if only part of the packet could
    be read; in this case, a subsequent call to recvpack will continue
    to read the same packet. Return 1 on success */
-int recvpack(int s, char **buf, phead_t **hdr)
+int recvpack(int s, char **buf, phead_t **hdr) {
+  return recvpack_t(s, buf, hdr, -1);
+}
+
+/* like recvpack, except it also has a timeout (in seconds). If t=-1,
+   no timeout. If t=0, return immediately if there is no data to read.
+   On timeout, return -4. */
+int recvpack_t(int s, char **buf, phead_t **hdr, int t)
 {
   int r, i;
   fd_set fs;
@@ -2386,13 +2588,29 @@ int recvpack(int s, char **buf, phead_t **hdr)
   static unsigned char *tdbuf = NULL;
   static int tpos;  /* how many characters in tdbuf */
 
+  /* if timeout requested, wait at most t seconds until data becomes
+     available for reading. */
+  if (t>=0) {
+    FD_ZERO(&fs);
+    FD_SET(s, &fs);
+    tv.tv_usec = 0;
+    tv.tv_sec = t;
+    
+    r = select(s+1, &fs, NULL, NULL, &tv);
+    if (r==-1) {
+      return -1;
+    } else if (r==0) { /* timeout */
+      return -4;
+    }
+  }
+
   if (!thdr)
   {
     unsigned char acc[4];
   
     for (i=0; i<4; i+=r) 
     {
-      r = read(s, acc+i, 4-i);
+      r = read(s, acc+i, 4-i);  /* note that this might hang */
       if (r <= 0)
       {
         *hdr = NULL;
@@ -2503,7 +2721,7 @@ int rsock(int s, char **buf)
   if (nvar("debug") == 2)
   {
     sock_t *sk = findsockfd(s); 
-    wp(wchan, ""DARK GREEN"<-- [from %d=%s] <%s>"WHITE"\n", s, sk ? sk->nm : "?", quote(*buf));
+    wp(wchan, ""DARK GREEN"<-- [from %d=%s] <%s>"WHITE"\n", s, sk ? sk->socknm : "?", quote(*buf));
     drw(wchan);
   }
   
