@@ -25,6 +25,7 @@
 #include <string.h>
 #include <math.h>
 #include <ctype.h>
+#include <fcntl.h>
 
 #include "getopt.h"
 #include "defines.h"
@@ -47,9 +48,11 @@
   #include "wind.h"
 #endif
 
+#ifdef MEMWATCH
+  #include "memwatch.h"
+#endif
 
 extern info_t info;
-extern scrls_t *mscroll, *scur;
 extern chans_t *chanl, *curchan, *recent;
 extern int tind, ircsock;
 extern int noprint, ircmode, lpbrk;
@@ -203,10 +206,11 @@ int icmpin(WINDOW *win, sock_t *m)
 
   line = getline(m->f);
 
-  if (!line) {
+  if (!line || line[0]==0) {
     fclose(m->f);
     delsock(m->fd);
     showresults(wchan, 2);
+    free(line);
     return(1);
   }
 
@@ -216,7 +220,7 @@ int icmpin(WINDOW *win, sock_t *m)
     for (cur=search; cur!=NULL; cur=cur->next) {
       cur->ping = -1;
     }
-
+    free(line);
     return(1);
   }
 
@@ -227,6 +231,8 @@ int icmpin(WINDOW *win, sock_t *m)
   *p = 0;
   p++;
   r = inet_aton(line, &addr);
+  free(line);
+
   if (!r) {  /* syntax error from napping, ignore */
     return(1);
   }
@@ -251,6 +257,7 @@ int icmperr(WINDOW *win, sock_t *m)
   if (line) {
     wp(win, "* Could not obtain ping results: %s\n", line);
     drw(win);
+    free(line);
   }
   
   fclose(m->f);
@@ -600,6 +607,7 @@ I_NAP_FUNC(sjchan)
   cur->p = 0;
   cur->next = NULL;
   curchan = cur;
+  indraw();    /* update command line with new channel */
   
   tin = -1;
 
@@ -858,9 +866,11 @@ I_NAP_FUNC(smpart)
     chanl = NULL;
     curchan = NULL;
   }
-  if (curchan && !strcmp(pt->nm, curchan->nm))
+  if (curchan && !strcmp(pt->nm, curchan->nm)) {
     upchan(curchan);
-  
+  }
+  indraw();
+
   while (pt->users)
     deluser(pt, pt->users->nm);
 
@@ -961,7 +971,7 @@ I_NAP_FUNC(sfreq)
      client to download any file, whether it was shared or not */
 
   if (info.shared_filename) {
-    libraryfile = info.shared_filename;
+    libraryfile = strdup(info.shared_filename);
   } else {
     libraryfile = home_file(LIBRARYFILE);
   }
@@ -971,9 +981,12 @@ I_NAP_FUNC(sfreq)
     wp(win, "%s* File \"%s\" was requested by %s, but is not shared!%s\n", RED, lfn, tok[0], WHITE);
     drw(win);
     free(lfn);
+    free(libraryfile);
     return(1);
   }
 
+  free(libraryfile);
+
   /* if /tquit is in effect, don't accept any new uploads */
   if (quit_after_transfers) {
     sendpack(s, F_GFR, "%s", str);
@@ -1013,7 +1026,7 @@ I_NAP_FUNC(sfreq)
        F_GFR here. Do the opennap servers react differently? ### check
        this another time. */
     sendpack(s, F_QLIMIT, "%s \"%s\" %i", tok[0], tok[1], r==2 ? nvar("maxuploads") : nvar("maxupuser"));
-    if (nvar("showtoomanyuploads")==1) {
+    if (nvar_default("showtoomanyuploads", 0)) {
       wp(win, "%s* Too many uploads %s(can't send \"%s\" to %s)%s\n", RED, r==2 ? "" : "for this user ", tok[1], tok[0], WHITE);
       drw(win);
     }
@@ -1056,7 +1069,7 @@ I_NAP_FUNC(ssf)
 {
   upload_t *task;
   sock_t *sk=NULL;
-  int k;
+  int k, r;
   struct sockaddr_in dst;
   
   /* find the corresponding upload */
@@ -1091,15 +1104,20 @@ I_NAP_FUNC(ssf)
   
   k = socket(AF_INET, SOCK_STREAM, 0);
   
-  /* we connect in a child process, so as not to hang */
+  /* use non-blocking i/o so that connect will not hang */
+  fcntl(k, F_SETFL, O_NONBLOCK);
 
-  if (fork()==0) {     /* child process */
-    connect(k, (struct sockaddr *)&dst, sizeof(dst));
-    exit(1);
+  r = connect(k, (struct sockaddr *)&dst, sizeof(dst));
+
+  if (r == -1 && errno != EINPROGRESS) {
+    wp(win, ""RED"* Error sending file (%s) to %s: %s"WHITE"\n", task->fn, task->nick, strerror(errno));
+    drw(win);
+    task->state = FAILED;
+    task->d_time = time(0);
+    close(k);
+    return(1);
   }
   
-  /* parent process */
-
   sk = addsock(k, gnum(1), S_R, initsend);
 
   /* update task */
@@ -1164,15 +1182,20 @@ I_NAP_FUNC(sget)
 
   k = socket(AF_INET, SOCK_STREAM, 0);
     
-  if (fork()==0)             /* child process */
-  {
-    r = connect(k, (struct sockaddr *)&task->addr, sizeof(task->addr));
-    /* if there is an error, we do not need to handle it here; 
-       it will be handled in initget. */
-    exit(1);
-  }
+  /* make socket non-blocking, so that connect will return immediately */
+  fcntl(k, F_SETFL, O_NONBLOCK);
   
-  /* parent process */
+  r = connect(k, (struct sockaddr *)&task->addr, sizeof(task->addr));
+
+  if (r == -1 && errno != EINPROGRESS) {
+    wp(win, ""RED"* Error getting file (%s) from %s: %s"WHITE"\n", fn, tok[0], strerror(errno));
+    drw(win);
+    free(task->check);
+    task->state = FAILED;
+    task->d_time = time(0);
+    close(k);
+    return(1);
+  }
 
   if (task->state == RRQUEUED) {
     wp(win, "* Getting \"%s\" from %s\n", task->fn, task->nick);
@@ -1440,7 +1463,7 @@ void showresults(WINDOW *win, int level) {
   
   /* show search result on main screen or result screen, depending
      on whether "noresultscreen" was selected. */
-  if (nvar("noresultscreen")==1) 
+  if (nvar_default("noresultscreen", 0)) 
   {
     if (!search) {
       return;
@@ -1507,7 +1530,7 @@ I_NAP_FUNC(srbrowse)
   t = ud_basename(tok[1]);
   
   cur->ping = -2;
-  cur->song = strdup(t);
+  cur->song = strdup(quote(t));
   cur->fn = strdup(tok[1]);
   cur->sz = atoi(tok[3]);
   cur->brate = atoi(tok[4]);
@@ -1557,7 +1580,7 @@ I_NAP_FUNC(sbrowse2err)
 I_NAP_FUNC(sbrowse2acc)
 {
   struct sockaddr_in dst;
-  int k;
+  int k, r;
   sock_t *sk=NULL;
   unsigned short int dstport;
   
@@ -1582,11 +1605,20 @@ I_NAP_FUNC(sbrowse2acc)
   dst.sin_family = AF_INET;
 
   k = socket(AF_INET, SOCK_STREAM, 0);
-  if (fork()==0) {     /* child process */
-    connect(k, (struct sockaddr *)&dst, sizeof(dst));
-    exit(1);
+
+  fcntl(k, F_SETFL, O_NONBLOCK);
+
+  r = connect(k, (struct sockaddr *)&dst, sizeof(dst));
+
+  if (r == -1 && errno != EINPROGRESS) {
+    wp(win, ""RED"* Error browsing %s: %s"WHITE"\n", tok[0], strerror(errno));
+    drw(win);
+    directbrowse.state = FAILED;
+    srch = 0;
+    close(k);
+    return(1);
   }
-  /* parent process */
+
   sk = addsock(k, "dirbrowse", S_R, initgetlist); /* initgetlist in event.c */
   
   directbrowse.state = CONNECTING;
@@ -1608,7 +1640,7 @@ I_NAP_FUNC(sbrowse2acc)
 I_NAP_FUNC(sbrowse2req)
 {
   struct sockaddr_in dst;
-  int k;
+  int k, r;
   sock_t *sk=NULL;
   unsigned short int dstport;
   
@@ -1641,11 +1673,20 @@ I_NAP_FUNC(sbrowse2req)
   dst.sin_family = AF_INET;
 
   k = socket(AF_INET, SOCK_STREAM, 0);
-  if (fork()==0) {     /* child process */
-    connect(k, (struct sockaddr *)&dst, sizeof(dst));
-    exit(1);
+
+  fcntl(k, F_SETFL, O_NONBLOCK);
+
+  r = connect(k, (struct sockaddr *)&dst, sizeof(dst));
+
+  if (r == -1 && errno != EINPROGRESS) {
+    if (nvar("debug") == 2) {
+      wp(win, ""RED"* Error sending list to %s: %s"WHITE"\n", tok[0], strerror(errno));
+      drw(win);
+    }
+    close(k);
+    return(1);
   }
-  /* parent process */
+
   sk = addsock(k, gnum(2), S_R, initsendlist); /* initsendlist in event.c */
 
   if (nvar("debug") == 2) {
@@ -1722,7 +1763,7 @@ I_NAP_FUNC(scping)
 
 I_NAP_FUNC(scpong)
 {
-  if (nvar("announcepongs") == 1)
+  if (nvar_default("announcepongs", 0))
   {
     wp(win, "%s* Received PONG from %s%s\n", RED, tok[0], WHITE);
     drw(win);
@@ -1770,7 +1811,7 @@ I_NAP_FUNC(sredir)
   sk = addsock(t, "server", S_R, inserv);
   
   curchan=NULL;
-  
+
   if (chanl)
   {
     for (cur=chanl;;cur=cur->next)
@@ -1780,6 +1821,7 @@ I_NAP_FUNC(sredir)
   }
   
   upchan(chanl);
+  indraw();
   
   return(1);
 }
@@ -1828,6 +1870,7 @@ I_NAP_FUNC(scycle)
   }
   
   upchan(chanl);
+  indraw();
   
   return(1);
 }
