@@ -28,6 +28,10 @@
 #include "lists.h"
 #include "scheck.h"
 
+#ifdef MEMWATCH
+  #include "memwatch.h"
+#endif
+
 extern info_t info;
 extern int ircmode, wmode;
 extern chans_t *curchan;    /* current channel, if any */
@@ -38,7 +42,7 @@ extern char *mnick;
 extern int noprint, lpbrk;
 extern int ircsock;
 extern alias_t *alhead;     /* alias list */
-extern out_nap_cmd_t out[]; /* list of builting commands */
+extern out_nap_cmd_t out[]; /* list of builtin commands */
 extern WINDOW *swin;
 
 /* the new model for managing input screens is as follows: */
@@ -74,14 +78,26 @@ int screen;
 /* curses windows and global state for main screen */
 WINDOW *wchan=NULL, *winput, *sep, *whead;
 int tind = 0;                 /* indentation of the title (channel topic) */
-scrls_t *mscroll, *scur;
 chans_t *recent = NULL;       /* most recent channel; (always NULL) */
 
-cmds_t *cmdl, *cmdlend = NULL; /* the command history. A doubly linked
-                                  list of lines, possibly empty. This
-                                  is read-only, except that new lines
-                                  can be appended at the end. */
-int cmdllen = 0;              /* length of the command history */
+scroll_t *mscroll = NULL;      /* top of the main scroll */
+scroll_t *mscrollend = NULL;   /* bottom of the main scroll */
+scroll_t *scur = NULL;         /* last line now displayed, or NULL for end */
+int mscrollsize = 0;           /* length of the main scroll */
+
+int lastlogflag = 0;           /* set to 1 during output of "lastlog"
+				  command - this causes these lines to
+				  be marked and ignored in future
+				  "lastlog" commands. */
+
+/* the command history */
+cmds_t *cmdl, *cmdlend = NULL; /* A doubly linked list of lines,
+				  possibly empty. This is read-only,
+				  except that new lines can be
+				  appended at the end, and the total
+				  length of the list is bounded by
+				  BACKLOG. */
+int cmdlsize = 0;              /* length of the command history */
 
 cmds_t *ccmd = NULL;          /* a pointer to the line in cmdl which
 				 the current edit is based on, or NULL if
@@ -465,8 +481,9 @@ int wp(WINDOW *win, const char *fmt, ...)
   if (win && !info.daemon)
   {
     addscroll(win, str);
-    if (nvar("noscroll")!=1)
+    if (nvar_default("noscroll", 0)==0) {
       scur = NULL;           /* be sure to scroll to bottom on output */
+    }
     dscr(win);
   }
   if (!win && info.daemon!=2) {
@@ -483,164 +500,251 @@ int wp(WINDOW *win, const char *fmt, ...)
   return(i);
 }
 
-scrls_t *scrend(scrls_t *t)
-{
-  scrls_t *cur, *cur1 = NULL;
-  
-  for (cur=t;cur!=NULL;cur=cur->next)
-    cur1 = cur;
-  
-  return(cur1);
+/* examines string s for escape sequences and returns a (statically
+   allocated) string which represents the escape sequences (color,
+   boldface) in effect at the end of string s. If end!=NULL, s is
+   considered to end there. */
+char *esc_getcolor(char *s, char *end) {
+  static char out[11] = BOLD;
+  char *col = out+strlen(BOLD);
+  int bf=0;
+
+  while (*s!=0 && (end==NULL || s<end)) {
+    if (*s == '\e') {
+      if (!strncmp(s, BOLD, strlen(BOLD))) {
+	bf = 1;
+	s += strlen(BOLD);
+      } else if (!strncmp(s, WHITE, strlen(WHITE))) {
+	bf = 0;
+	col[0] = 0;
+	s += strlen(WHITE);
+      } else if (s[3] == 'm') {
+	strncpy(col, s, 4);
+	col[4] = 0;
+	s += 4;
+      } else {
+	strncpy(col, s, 5);
+	col[5] = 0;
+	s += 5;
+      }
+    } else {
+      s++;
+    }
+  }
+  if (bf) {
+    return out;
+  } else {
+    return col;
+  }
 }
 
+/* calculate the "printable" length of a string, not including any
+   escape sequences */
+int esc_getlength(char *s) {
+  int len = 0;
+
+  while (*s) {
+    if (*s == '\e' && s[3] == 'm') {
+      s+=4;
+    } else if (*s == '\e') {
+      s+=5;
+    } else {
+      s++;
+      len++;
+    }
+  }
+  return len;
+}
+
+/* NOTE: the scroll log is the biggest memory hog in nap, as it can
+   grow to thousands of lines. Therefore some special care was taken
+   to allocate no more room for each line than necessary. In addition,
+   we limit the size of the scroll to scrollsize. */
+
 /* add a line to the "scroll" list of the main screen */
 void addscroll(WINDOW *win, char *buf)
 {
-  int i, j, k = 0, m = 0, bf = 0;
-  scrls_t *cur, *cur1, *par;
-  char col[10];
-  
-  cur = scrend(mscroll);
-  cur1 = cur;
+  scroll_t *cur, *new;
+  char *p, *q;
+  char *linebreak, *linebreak1=NULL;
+  int scrollsize;
 
-  if (cur1 == NULL)
-  {
-    cur = (scrls_t *)malloc(sizeof(scrls_t));
-    cur->prev = cur1;
-    cur->next = NULL;
-    cur->ddd = 0;
-    cur->chan = recent;
-    cur->own = NULL;
-    memset(cur->ln, 0, sizeof(cur->ln));
-    mscroll = cur;
-  }
-  else if (cur1->ddd)
-  {
-    cur = (scrls_t *)malloc(sizeof(scrls_t));
-    cur->prev = cur1;
-    cur->next = NULL;
-    cur->ddd = 0;
-    cur->chan = recent;
-    cur->own = NULL;
-    memset(cur->ln, 0, sizeof(cur->ln));
-    cur1->next = cur;
-  }
+  /* len: length of the current line to append to, or -1 if the last
+     character was '\n' and a line has not yet been created. Note
+     len>=0 implies mscrollend != NULL */
+  static int len = -1;   
+  static int size, asize;  /* real size and allocated size of current
+                              line, if len != -1. */
+
+  /* process each character or espace sequence from buf */
   
-  if (!cur->ddd)
-  {
-    for (;cur->ln[k];k++)
-    {
-      if (cur->ln[k] == '\e')
-      {
-        if (!strncmp(cur->ln+k, BOLD, strlen(BOLD)))
-          bf = 1;
-        else if (!strncmp(cur->ln+k, WHITE, strlen(WHITE)))
-          bf = 0;
-        if (cur->ln[k+3] == 'm')
-          m+=4;
-        else
-          m+=5;
-      }
+  for (p=buf; *p!=0; p++) {
+
+    if (*p == '\r') {    /* ignore - concession to MSDOS? */
+      continue;
     }
-  }
-  
-  par = cur;
-  
-  memset(col, 0, sizeof(col));
-  
-  for (i=0,j=strlen(buf);i<j;i++,k++)
-  {
-    if (buf[i] == '\r')
-      i++;
-    if (k >= (COLS+m-1) || buf[i] == '\n')
-    {
-      if (k == (COLS+m-1) && strchr(cur->ln, ' '))
-      {
-        for (;!isspace(cur->ln[k])&&k>=0;k--,i--);
-        i++;
-      }
-      cur->ln[k] = '\0';
-      strcat(cur->ln, WHITE);
-      cur->ddd = 1;
-      k = 0;
-      m = 0;
-      if (buf[i] == '\n')
-        i++;
-      if (i == j)
-        break;
-      cur1 = cur;
-      cur = (scrls_t *)malloc(sizeof(scrls_t));
-      cur->prev = cur1;
-      cur->next = NULL;
-      cur->ddd = 0;
+    
+    /* create a new logical line if there is no current line (previous
+       character was '\n'). */
+    
+    if (len==-1) {
+      /* create a new empty line */
+      cur = (scroll_t *)malloc(sizeof(scroll_t));
+      cur->lastlog = lastlogflag;
       cur->chan = recent;
-      cur1->next = cur;
-      if (buf[i-1] == '\n')
-      {
-        cur->own = NULL;
-        par = cur;
-        bf = 0;
-      }
-      else
-        cur->own = par;
-      memset(cur->ln, 0, sizeof(cur->ln));
-      if (*col)
-      {
-        if (bf && strcmp(col, BOLD))
-          strcpy(cur->ln, BOLD);
-        strcat(cur->ln, col);
-        m = k = strlen(cur->ln);
-      }
-      else if (bf)
-      {
-        strcpy(cur->ln, BOLD);
-        m = k = strlen(cur->ln);
-      }
-      if (cur->own)
-        m-=TABSIZE;
+      cur->own = NULL;
+      asize = 256;
+      cur->line = (char *)malloc(asize);
+      cur->line[0] = 0;
+
+      /* attach to scroll */
+      dlist_append(mscroll, mscrollend, cur);
+      mscrollsize++;
+      
+      /* initialize local state for appending to this line */
+      len = 0;
+      size = 0;
     }
-    if (buf[i] == '\e')
-    {
-      if (buf[i+3] == 'm')
-      {
-        if (!strncmp(buf+i, BOLD, strlen(BOLD)))
-          bf = 1;
-        else if (!strncmp(buf+i, WHITE, strlen(WHITE)))
-          bf = 0;
-        memset(col, 0, sizeof(col));
-        strncpy(col, buf+i, 4);
-        strncpy(cur->ln+k, buf+i, 4);
-        m+=4;
-        i+=3;
-        k+=3;
+
+    cur = mscrollend;
+
+    /* make sure at least 5 more characters fit in current line */
+    if (size+5+1 > asize) {
+      asize += 256;
+      cur->line = (char *)realloc(cur->line, asize);
+    }
+    
+    if (*p == '\n') {
+      /* mark this line ended - next printable character will start a
+         new (logical and physical) line */
+      q = esc_getcolor(cur->line, NULL);
+      if (q[0] != 0) {
+ 	strcat(cur->line, WHITE);
+	size += strlen(WHITE);
       }
-      else
-      {
-        memset(col, 0, sizeof(col));
-        strncpy(col, buf+i, 5);
-        strncpy(cur->ln+k, buf+i, 5);
-        m+=5;
-        i+=4;
-        k+=4;
+      len = -1;
+      /* free some memory by allocating a fixed size for this line */
+      q = strdup(cur->line);
+      free(cur->line);
+      cur->line = q;
+
+      continue;
+
+    } else if (*p == '\e') {
+      /* escape sequence - copy to current line without counting its
+         length towards the total */
+      if (p[3] == 'm') {
+	strncat(cur->line, p, 4);
+	size += 4;
+	p += 3;
+      } else {
+	strncat(cur->line, p, 5);
+	size += 5;
+	p += 4;
       }
+      continue;
+      
+    } else {   /* printable character - copy it to current line */
+      if (*p!=' ' && len >= COLS-1) {
+	/* the character will not fit. Find a good place to break line */
+	linebreak = strrchr(cur->line, ' ');
+	if (linebreak - cur->line < COLS/2 || linebreak - cur->line < 40) {
+	  linebreak = NULL;
+	}
+	if (linebreak) {
+	  linebreak1 = linebreak + 1;
+	  while (linebreak > cur->line && linebreak[-1]==' ') {
+	    linebreak--;
+	  }
+	}
+	
+	/* create continuation line */
+	new = (scroll_t *)malloc(sizeof(scroll_t));
+	new->lastlog = lastlogflag;
+	new->chan = recent;
+	new->own = cur->own ? cur->own : cur;
+	asize = 256;
+	new->line = (char *)malloc(asize);
+	new->line[0] = 0;
+
+	/* copy over stuff from old line to new one */
+	q = esc_getcolor(cur->line, linebreak);
+	strcat(new->line, q);
+	strcat(new->line, LINEINDENT);
+	if (linebreak) {
+	  strcat(new->line, linebreak1);
+	}
+	len = esc_getlength(new->line);
+	size = strlen(new->line);
+	
+	/* truncate old line if necessary */
+	if (linebreak) {
+	  *linebreak = 0;
+	}
+
+	/* re-allocate old line using smallest amount of space */
+	q = strdup(cur->line);
+	free(cur->line);
+	cur->line = q;
+
+	/* attach new line to scroll */
+	dlist_append(mscroll, mscrollend, new);
+	mscrollsize++;
+
+	cur = new;
+      }
+      
+      /* character fits; append it */
+      strncat(cur->line, p, 1);
+      len++;
     }
-    else
-      cur->ln[k] = buf[i];
   }
+  
+  scrollsize = nvar_default("scrollsize", SCROLLSIZE);
+
+  while (scrollsize > 0 && mscrollsize > scrollsize) { 
+    /* limit the size of mscroll */
+    do {
+      dlist_unlink_first(mscroll, mscrollend, cur);
+      mscrollsize--;
+      free(cur->line);
+      free(cur);
+    } while (mscroll && mscroll->own);
+  }
+}
+
+/* clear the main scroll */
+void clearscroll(void) {
+  scroll_t *elt;
+  
+  dlist_forall_unlink(elt, mscroll, mscrollend) {
+    free(elt->line);
+    free(elt);
+  }
+  mscrollsize = 0;
+  scur = NULL;
+  dscr(wchan);
+}  
+
+/* scroll to bottom */
+void scrollbottom(void) {
+  scur = NULL;
+  dscr(wchan);
 }
 
 void dscr(WINDOW *win)
 {
-  scrls_t *cur;
+  scroll_t *cur;
   int i, j, k, m, at=0, t, c;
   unsigned char bf = 0, uf = 0, gf = 0;
   char ebuf[6];
+  int col; /* column */
   
   werase(win);
   wmove(win, 0, 0);
   
   if (scur == NULL)
-    cur = scrend(mscroll);
+    cur = mscrollend;
   else
     cur = scur;
   
@@ -664,28 +768,27 @@ void dscr(WINDOW *win)
   
   c = win->_maxy;
   
-  for (j=0,i=0;i<=c;cur=cur->next)
+  for (j=0,i=0; i<=c; cur=cur->next)
   {
     if (cur == NULL)
       break;
     if (cur->chan && (wmode && cur->chan != curchan))
       continue;
-    if (cur->own)
-      wmove(win, j, TABSIZE);
-    else
+    if (!cur->own)
     {
       bf = 0;
       uf = 0;
       gf = 0;
-      wmove(win, j, 0);
     }
+    wmove(win, j, 0);
+    col = 0;
 #ifndef MCURSES
-    for (k=0;cur->ln[k];k++)
+    for (k=0; cur->line[k] && col < COLS; k++)
     {
-      if (cur->ln[k] == '\e' && cur->ln[k+1] == '[' && isdigit(cur->ln[k+2]))
+      if (cur->line[k] == '\e' && cur->line[k+1] == '[' && isdigit(cur->line[k+2]))
       {
-        for (m=0; cur->ln[k] && cur->ln[k]!='m'; m++, k++)
-          ebuf[m] = cur->ln[k];
+        for (m=0; cur->line[k] && cur->line[k]!='m'; m++, k++)
+          ebuf[m] = cur->line[k];
         ebuf[m] = 'm';
         ebuf[m+1] = '\0';
         t = doesc(win, ebuf);
@@ -694,7 +797,7 @@ void dscr(WINDOW *win)
         else
           at |= t;
       }
-      else if (cur->ln[k] == 2)
+      else if (cur->line[k] == 2)
       {
         if (!bf)
         {
@@ -707,7 +810,7 @@ void dscr(WINDOW *win)
           bf = 0;
         }
       }
-      else if (cur->ln[k] == 31)
+      else if (cur->line[k] == 31)
       {
         if (!uf)
         {
@@ -720,32 +823,35 @@ void dscr(WINDOW *win)
           uf = 0;
         }
       }
-      else if (cur->ln[k] == 7)
+      else if (cur->line[k] == 7)
       {
         if (!gf && i == c)
           beep();
         gf = 1;
       }
-      else if (cur->ln[k] == 15)
+      else if (cur->line[k] == 15)
         uf = bf = at = 0;
-      else
-        waddch(win, cur->ln[k]|at);
+      else {
+        waddch(win, cur->line[k]|at);
+	col++;
+      }
     }
 #else
-    waddstr(win, cur->ln);
+    waddstr(win, cur->line);
 #endif
     i++;
     j++;
   }
+  scur = cur ? cur->prev : NULL;
 }
 
 void dscroll(WINDOW *win, int n)
 {
   int i;
-  scrls_t *cur;
+  scroll_t *cur;
   
   if (scur == NULL)
-    cur = scrend(mscroll);
+    cur = mscrollend;
   else
     cur = scur;
   
@@ -858,16 +964,16 @@ int input(WINDOW *win, sock_t *m)
 	cmdl=elt;
       }
       cmdlend=elt;
-      cmdllen++;
+      cmdlsize++;
       
       /* make sure command history does not get too long */
-      if (cmdllen > BACKLOG) {
+      if (cmdlsize > BACKLOG) {
 	elt = cmdl->next;
 	free(cmdl->cmd);
 	free(cmdl);
 	cmdl = elt;
 	cmdl->prev = NULL;
-	cmdllen--;
+	cmdlsize--;
       }
     }
 
@@ -1107,7 +1213,7 @@ int input(WINDOW *win, sock_t *m)
 
   case KEY_PPAGE:                       /* PgUp */
   case 16:                              /* Ctrl-P */ 
-    dscroll(win, 10);
+    dscroll(win, (LINES-2)/2);
     indraw();
     ebuf = 0;
     break;
@@ -1115,7 +1221,7 @@ int input(WINDOW *win, sock_t *m)
   case KEY_NPAGE:                       /* PgDn */
   case 14:                              /* Ctrl-N */
   case 22:                              /* Crtl-V */
-    dscroll(win, -10);
+    dscroll(win, -(LINES-2)/2);
     indraw();
     ebuf = 0;
     break;
